//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DumpRom.C
//
//		Uses the TILynx 1394 API to dump the configuration ROM of a specified
//		1394 peripheral.
//
//		Use PCI Lynx 2.1 kit availible on the TI web site.
//
//		Set up the directories as follows:
//
//		..\inc	copy 1394api.h and 1394type.h from PCI Lynx kit 2.1
//		..\lib	copy cls1394.lib from PCI Lynx kit 2.1
//		..\bin	copy cls1394.dll from PCI Lynx kit 2.1
//				we also copy the dumprom.exe file there for ease of execution
//				and debugging.
//
//		Developed using MSVC 5.0.
//
//	Bill Alexander
//	10/13/98
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define _APP_MAIN_MODULE_
#define _DUMPROM_C_

#define STRICT
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <windows.h>

//
// The 1394api.h header file is provided by Texas Instruments to be
// with their Windows 95 API software.  The software is available on
// TI's website: http://www.ti.com
//
#include "..\inc\1394api.h"

// My local include files
#include "dbg.h"
#include "dumprom.h"

#pragma pack()

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Main
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int main(int argc, char *argv[])
{
	BOOL BusReset;
	BOOL ccode;

	// Program Information Structure
	ProgInfoT pi;

	MyPrintf(&pi, VL_MAJOR_STEP, "1394 Dump ROM Utility\n");

	// Init vars
	pi.hHost = 0;
	pi.hDev = 0;
	pi.DriverInitialized = FALSE;
	pi.DeviceId = DEFAULT_DEVICE_GUID;

	// Process the command line
	ccode = ProcessCommandLine (argc, argv, &pi);
	if (ccode != TRUE)
		goto CommandLineError;

	// Initialize 1394 driver and DLL and reset the bus.
	MyPrintf (&pi, VL_MAJOR_STEP, "Initializing 1394 drivers ... ");
	BusReset = TRUE;
	ccode = cls1394Initialize (&BusReset);
	if (ccode != TRUE)
		goto DriverInitError;
	MyPrintf (&pi, VL_MAJOR_STEP, "OK\n");
	pi.DriverInitialized = TRUE;

	// Open the host 1394 device
	MyPrintf (&pi, VL_MAJOR_STEP, "Opening Host Device ... ");
	ccode = HostOpen (&pi.hHost);
	if (ccode != TRUE)
		goto HostOpenError;
	MyPrintf (&pi, VL_MAJOR_STEP, "OK\n");

	// Open the 1394 device
	MyPrintf1 (&pi, VL_MAJOR_STEP, "Opening 1394 Device %08X ... ", pi.DeviceId);
	ccode = DeviceOpen (pi.DeviceId, &pi.hDev);
	if (ccode != TRUE)
		goto DevOpenError;
	MyPrintf (&pi, VL_MAJOR_STEP, "OK\n");

	// Dump Config ROM
	ccode = DumpConfigRom (&pi);
	if (ccode != TRUE)
		goto Exit;

	// If using a specified offset, we don't need to do anything else
	if (pi.UseSpecifiedOffset == TRUE)
		goto Exit;

Exit:

	// Close connection to host board
	if (pi.hHost != 0)
	{
		MyPrintf (&pi, VL_MAJOR_STEP, "Closing host device ... ");
		ccode = cls1394CloseHandle (pi.hHost);
		pi.hHost = 0;
		if (ccode != TRUE)
			goto DevCloseError;
		MyPrintf (&pi, VL_MAJOR_STEP, "OK\n");
	}

	// Close connection to 1394 device
	if (pi.hDev != 0)
	{
		MyPrintf1 (&pi, VL_MAJOR_STEP, "Closing device %08X... ", pi.DeviceId);
		ccode = cls1394CloseHandle (pi.hDev);
		pi.hDev = 0;
		if (ccode != TRUE)
			goto DevCloseError;
		MyPrintf (&pi, VL_MAJOR_STEP, "OK\n");
	}

	// Unload the drivers
	if (pi.DriverInitialized == TRUE)
	{
		MyPrintf (&pi, VL_MAJOR_STEP, "Unloading drivers ...");
		ccode = cls1394Terminate ();
		pi.DriverInitialized = FALSE;
		if (ccode != TRUE)
			goto DriverTerminateError;
		MyPrintf (&pi, VL_MAJOR_STEP, "OK\n");
	}

	// Pause before exiting
	if (pi.StopBeforeExit == TRUE)
	{
		MyPrintf (&pi, VL_ALWAYS, "Hit any key to continue.\n");
		getch();
	}
	
	return (0);

CommandLineError:
	MyPrintf(&pi, VL_ERROR, "\nERROR:  Command line syntax error\n\n");
	Usage (&pi);
	goto Exit;

DriverInitError:
	MyPrintf(&pi, VL_ERROR, "\nError: Cannot initialize 1394 drivers\n\n");
	goto Exit;

HostOpenError:
	MyPrintf(&pi, VL_ERROR, "\nError: Host device not found\n\n");
	goto Exit;

DevOpenError:
	MyPrintf1(&pi, VL_ERROR, "\nError: Device %08X not found\n\n", pi.DeviceId);
	goto Exit;

DevCloseError:
	MyPrintf1(&pi, VL_ERROR, "\nError: Device %08X cannot be closed properly\n\n", pi.DeviceId);
	goto Exit;

DriverTerminateError:
	MyPrintf(&pi, VL_ERROR, "\nError: Cannot terminate 1394 drivers\n\n");
	goto Exit;

}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ProcessCommandLine
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL ProcessCommandLine (int argc, char *argv[], PProgInfoT pi)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	int i;
	int ccode;

	// Assume failure
	retcode = FALSE;

	// Set default values for command line options
	pi->StopBeforeExit = FALSE;
	pi->VerbosityLevel = 0;
	pi->UseSpecifiedOffset = FALSE;
	pi->SpecifiedOffset = 0;

	// Run through all of the command line options
	// NOTE:  skip the first command line arg.
	// This is the program full path.
	for (i=1; (i<argc) && (*argv[i] != '\0'); i++)
	{
		char *p;
		char c;
		p = argv[i];

		// Parse current arg
		switch (*p)
		{
			// Parse Option	
			case '-':
			case '/':

				// Advance to option
				p++;

				// Parse option character
				switch (toupper(*p++))
				{
					// Verosity Setting
					case 'V':

						// Verify the level number
						sscanf (p, "%c", &c);
						if (c > '9' || c < '0')
							goto Exit;

						// Set the level
						sscanf (p++, "%d", &pi->VerbosityLevel);
						DBOUT1 (LF_INTERNAL, "pi->VerbosityLevel = %08X\r\n", pi->VerbosityLevel);

						// Check if end of option
						if (*p != '\0')
							goto Exit;

						break;

					// Stop Before Exit Option
					case 'S':

						// Set the option
						pi->StopBeforeExit = TRUE;

						// Check if end of option
						if (*p != '\0')
							goto Exit;

						break;

					// Offset 
					case 'O':

						// Extract the 4 digit specified offset
						pi->UseSpecifiedOffset = TRUE;
						ccode = sscanf (p, "%04X", &pi->SpecifiedOffset);
						if (ccode == 0)
							goto Exit;
						DBOUT1 (LF_INTERNAL, "pi->SpecifiedOffset = %08X\r\n", pi->SpecifiedOffset);

						break;

					// Device ID
					case 'D':

						// Extract the 8 digit Device ID
						ccode = sscanf (p, "%08X", &pi->DeviceId);
						if (ccode == 0)
							goto Exit;
						DBOUT1 (LF_INTERNAL, "pi->DeviceId = %08X\r\n", pi->DeviceId);

						break;

					default:
						goto Exit;
						break;
				}

				break;

			// There's a problem
			default:
				goto Exit;
				break;
		}
	}

	// OK
	retcode = TRUE;

Exit:

	return (retcode);

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Usage
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void Usage (PProgInfoT pi)
{
	MyPrintf (pi, VL_ERROR, "Usage:\n");
	MyPrintf (pi, VL_ERROR, "\n");
	MyPrintf (pi, VL_ERROR, "DumpRom [-d<device id>] [-v<0-9>] [-s]\n");
	MyPrintf (pi, VL_ERROR, "\n");
	MyPrintf (pi, VL_ERROR, "     -d<device id>   32-bit hexidecimal device ID \n");
	MyPrintf (pi, VL_ERROR, "                     Default: 0x00A0B800\n");
	MyPrintf (pi, VL_ERROR, "\n");
	MyPrintf (pi, VL_ERROR, "     -v<0-9>         Set verbosity level\n");
	MyPrintf (pi, VL_ERROR, "                     0 = none\n");
	MyPrintf (pi, VL_ERROR, "                     1 = show major steps\n");
	MyPrintf (pi, VL_ERROR, "                     2 = show major steps and debug information\n");
	MyPrintf (pi, VL_ERROR, "\n");
	MyPrintf (pi, VL_ERROR, "     -o<1394 offset> Dump only the quadlet at the specified 4 digit\n");
	MyPrintf (pi, VL_ERROR, "                     configuration rom offset.  Offset added to base\n");
	MyPrintf (pi, VL_ERROR, "                     of 0x0000FFFFF0000400.\n");
	MyPrintf (pi, VL_ERROR, "                     Default: dumps the whole configuration rom\n");
	MyPrintf (pi, VL_ERROR, "\n");
	MyPrintf (pi, VL_ERROR, "     -s              Stop before exiting\n");
	MyPrintf (pi, VL_ERROR, "\n");
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MyPrintf
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void MyPrintf (PProgInfoT pi, int vl, char *fmt)
{
	//#define DEBUGFLAGS DebugThisRoutine
	if (vl <= pi->VerbosityLevel)
	{
		DBOUT (LF_INTERNAL, fmt);
		printf (fmt);
	}
	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MyPrintf1
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void MyPrintf1 (PProgInfoT pi, int vl, char *fmt, DWORD p1)
{
	char msg[500];
	sprintf (msg, fmt, p1);
	MyPrintf (pi, vl, msg);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MyPrintf2
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void MyPrintf2 (PProgInfoT pi, int vl, char *fmt, DWORD p1, DWORD p2)
{
	char msg[500];
	sprintf (msg, fmt, p1, p2);
	MyPrintf (pi, vl, msg);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// WriteBuffer
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL WriteBuffer (
			cls1394HANDLE hDev,
			LARGE_INTEGER liAddress,
			PQUADLET Buffer,
			DWORD BufferSize)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	BOOL ccode;
	clsAsyncWrite aw;
	int i;

	DBOUT (LF_ENTRY, "WriteBuffer\r\n");
	DBOUT1 (LF_ENTRY_PARAMS, "hDev = %08X\r\n", hDev);
	DBOUT2 (LF_ENTRY_PARAMS, "liAddress = %08X%08X\r\n", liAddress.HighPart, liAddress.LowPart);
	DBOUT1 (LF_ENTRY_PARAMS, "Buffer = %08X\r\n", Buffer);
	DBOUT1 (LF_ENTRY_PARAMS, "BufferSize = %08X\r\n", BufferSize);

	// Assume failure
	retcode = FALSE;

	// Setup async write structure
	aw.DestinationAddress = &liAddress;
	aw.nNumberOfBytesToWrite = BufferSize;
	aw.nBlockSize = 0;
	aw.fulFlags = 0;
	aw.lpBuffer = Buffer;

	// Call the driver to do the async write
	ccode = cls1394DeviceIoControl(
				hDev,
				CLS_REQUEST_ASYNC_WRITE,
				&aw,
				sizeof(clsAsyncWrite),
				0,
				0,
				0,
				0);
	if (ccode != TRUE)
		goto Exit;

	// Delay for a little while
	for(i=0; i<20; i++)
	{
		ctMicrosec(1);
	}

	// OK
	retcode = TRUE;

Exit:

	DBOUT1 (LF_EXIT, "WriteBuffer, retcode = %08X\r\n", retcode);
	return (retcode);

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// WriteQuadlet
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL WriteQuadlet (
			cls1394HANDLE hDev,
			LARGE_INTEGER liAddress,
			QUADLET qValue)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	BOOL ccode;
	clsAsyncWrite aw;
	int i;

	DBOUT (LF_ENTRY, "WriteQuadlet\r\n");
	DBOUT2 (LF_ENTRY_PARAMS, "liAddress = %08X%08X\r\n", liAddress.HighPart, liAddress.LowPart);
	DBOUT1 (LF_ENTRY_PARAMS, "qValue = %08X\r\n", qValue);

	// Assume failure
	retcode = FALSE;

	// Setup async write structure
	aw.DestinationAddress = &liAddress;
	aw.nNumberOfBytesToWrite = sizeof(QUADLET);
	aw.nBlockSize = 0;
	aw.fulFlags = 0;
	aw.lpBuffer = &qValue;

	// Call the driver to do the async write
	ccode = cls1394DeviceIoControl(
				hDev,
				CLS_REQUEST_ASYNC_WRITE,
				&aw,
				sizeof(clsAsyncWrite),
				0,
				0,
				0,
				0);
	if (ccode != TRUE)
		goto Exit;

	// Delay for a little while
	for(i=0; i<20; i++)
	{
		ctMicrosec(1);
	}

	// OK
	retcode = TRUE;

Exit:

	DBOUT1 (LF_EXIT, "WriteQuadlet, retcode = %08X\r\n", retcode);
	return (retcode);

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// WriteVerifyQuadlet
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL WriteVerifyQuadlet (
			cls1394HANDLE hDev,
			LARGE_INTEGER liAddress,
			QUADLET qValue,
			DWORD to)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	BOOL ccode;
	QUADLET qVerify;
	DWORD tc;

	DBOUT (LF_ENTRY, "WriteVerifyQuadlet\r\n");
	DBOUT2 (LF_ENTRY_PARAMS, "liAddress = %08X%08X\r\n", liAddress.HighPart, liAddress.LowPart);
	DBOUT1 (LF_ENTRY_PARAMS, "qValue = %08X\r\n", qValue);

	// Assume failure
	retcode = FALSE;

	// Write the quadlet
	ccode = WriteQuadlet (hDev, liAddress, qValue);
	if (ccode != TRUE)
		goto Exit;

	// Verify the quadlet
	// Wait for the completion with a timeout
	tc = GetTickCount();
	do
	{
		// Read the quadlet
		qVerify = ~qValue;
		ccode = ReadQuadlet (hDev, liAddress, &qVerify);
		if (ccode != TRUE)
			continue;

		// Verify the QUADLET
		if (qValue == qVerify)
			goto OK;

	} while ((tc + to) > GetTickCount());

	// Timed out -- FAILED!
	goto Exit;

OK:

	// OK
	retcode = TRUE;

Exit:

	DBOUT1 (LF_EXIT, "WriteVerifyQuadlet, retcode = %08X\r\n", retcode);
	return (retcode);

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ReadQuadlet
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL ReadQuadlet(cls1394HANDLE hDev, LARGE_INTEGER liAddress, QUADLET *qValue)
{
    clsAsyncRead aw;
    BOOL rc;
    int i;

    aw.DestinationAddress = &liAddress;
    aw.nNumberOfBytesToRead = sizeof(QUADLET);
    aw.nBlockSize = 0;
    aw.fulFlags = 0;
    aw.lpBuffer = qValue;

    rc = cls1394DeviceIoControl(hDev,
                                CLS_REQUEST_ASYNC_READ,
                                &aw,
                                sizeof(clsAsyncRead),
                                0,
                                0,
                                0,
                                0);

    for(i=0; i<20; i++)
    {
        ctMicrosec(1);
    }

    return(rc);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// HostOpen
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL HostOpen (cls1394HANDLE *hHost)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	cls1394HANDLE lochHost;
	DWORD i;

	DBOUT (LF_ENTRY, "HostOpen\r\n");

	// Assume failure
	retcode = FALSE;
	lochHost = 0;

	// Try to open each Host GUID in turn
	for (i=0; i<sizeof(HostGuids)/sizeof(DWORD); i++)
	{
		// Try to open the current GUID
		DBOUT1 (LF_INTERNAL, "Opening Host with GUID = %08X\r\n", HostGuids[i]);
		lochHost = cls1394CreateFile (HostGuids[i], 0);
		if (lochHost)
			goto OK;
	}

	// FAILED!
	goto Exit;

OK:

	// OK
	retcode = TRUE;
	*hHost = lochHost;
	DBOUT1 (LF_EXIT_PARAMS, "*hHost = %08X\r\n", *hHost);

Exit:

	DBOUT1 (LF_EXIT, "HostOpen, retcode = %08X\r\n", retcode);
	return (retcode);

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// HostReadNodeId
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL HostReadNodeId (cls1394HANDLE hHost, QUADLET *qValue)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	BOOL ccode;

	DBOUT (LF_ENTRY, "HostOpen\r\n");

	// Assume failure
	retcode = FALSE;

	// Read the host's node id
	ccode = ReadQuadlet (hHost, NODE_IDS, qValue);
	if (ccode != TRUE)
		goto Exit;

	// OK
	retcode = TRUE;
	DBOUT1 (LF_EXIT_PARAMS, "*qValue = %08X\r\n", *qValue);

Exit:

	DBOUT1 (LF_EXIT, "HostOpen, retcode = %08X\r\n", retcode);
	return (retcode);

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DeviceOpen
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL DeviceOpen (DWORD DeviceId, cls1394HANDLE *hDev)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	cls1394HANDLE lochDev;
	DWORD j;

	DBOUT (LF_ENTRY, "DeviceOpen\r\n");

	// Assume failure
	retcode = FALSE;
	lochDev = 0;

	// Try the 256 combinations of sub dev types
	for (j=0; j<=0xFF; j++)
	{
		// Try to open the current GUID
		DBOUT1 (LF_INTERNAL, "Opening device with GUID = %08X\r\n", DeviceId);
		lochDev = cls1394CreateFile (DeviceId+j, 1);
		if (lochDev)
			goto OK;
	}

	// FAILED!
	goto Exit;

OK:

	// OK
	retcode = TRUE;
	*hDev = lochDev;
	DBOUT1 (LF_EXIT_PARAMS, "*hDev = %08X\r\n", *hDev);

Exit:

	DBOUT1 (LF_EXIT, "DeviceOpen, retcode = %08X\r\n", retcode);
	return (retcode);

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DumpConfigRom
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL DumpConfigRom (PProgInfoT pi)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	LARGE_INTEGER liAddress;
	QUADLET qValue;
	int i;
	BOOL ccode;

	DBOUT (LF_ENTRY, "DumpConfigRom\r\n");
	DBOUT1 (LF_ENTRY_PARAMS, "pi = %08X\r\n", pi);

	// Assume failure
	retcode = FALSE;

	// Read and print the config rom
	if (pi->UseSpecifiedOffset == TRUE)
	{
		// Read and printf the quadlet at the specified cfg rom offset only
		qValue = 0;
		liAddress = CONFIG_ROM;
		liAddress.LowPart += pi->SpecifiedOffset;
		ccode = ReadQuadlet(pi->hDev, liAddress, &qValue);
		if (ccode != TRUE)
			goto ReadQuadError;
		MyPrintf2(pi, VL_ALWAYS, "\nQuadlet at offset %04X = %08X\n", pi->SpecifiedOffset, qValue);
	}
	else
	{
		// Read and print each quadlet in the Config ROM space
		// until we get an error back
		liAddress = CONFIG_ROM;		// Cfg ROM - info len, CRC len, RIN CRC
		i = 0;
		MyPrintf (pi, VL_ALWAYS, "\n");
		MyPrintf1(pi, VL_ALWAYS, "Offset Data\n", qValue);
		MyPrintf (pi, VL_ALWAYS, "------ -------- -------- -------- -------- -------- -------- -------- --------");
		do
		{
			// Read the next quadlet
			qValue = 0;
			ccode = ReadQuadlet(pi->hDev, liAddress, &qValue);
			if (ccode != TRUE)
				break;

			// Display the quadlet
			if ((i % 8) == 0)
			{
				MyPrintf1 (pi, VL_ALWAYS, "\n%6d ", i*4);
			}
			MyPrintf1 (pi, VL_ALWAYS, "%08X ", qValue);

			// Advance to next quadlet
			i++;
			liAddress.LowPart = liAddress.LowPart + 4;

		} while (1);
		MyPrintf (pi, VL_ALWAYS, "\n");
	}

	// OK
	retcode = TRUE;

Exit:

	DBOUT1 (LF_EXIT, "DumpConfigRom, retcode = %08X\r\n", retcode);
	return (retcode);

ReadQuadError:
	MyPrintf2(pi, VL_ERROR, "\nERROR:  Cannot read the quadlet at address %08X%08X\n", liAddress.HighPart, liAddress.LowPart);
	goto Exit;

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// LI_HOST_SWAP
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
LARGE_INTEGER LI_HOST_SWAP (LARGE_INTEGER li)
{
	LARGE_INTEGER retcode;
	retcode.HighPart = li.LowPart;
	retcode.LowPart = li.HighPart;
	return (retcode);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// HostSwapQuadlet
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void HostSwapQuadlet (QUADLET qValue, PQUADLET dst)
{
	#ifdef I386

		QUADLET tmp;

		tmp = qValue;
		_asm
		{
			mov	eax, tmp;
			xchg	al, ah;
			clc;
			ror	eax, 16;
			xchg	al, ah;
			mov	tmp, eax
		}
		*dst = tmp;

	#else

		*dst = qValue;

	#endif
}


